@@ -1,9 +1,9 @@
 /*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@vu.nl
-    WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2022, University of Amsterdam
+    E-mail:        jan@swi-prolog.org
+    WWW:           https://www.swi-prolog.org
+    Copyright (c)  2011-2025, University of Amsterdam
 			      VU University Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -97,7 +97,7 @@ get_trace_store(int create)
     { btrace *s = malloc(sizeof(*s));
       if ( s )
       { memset(s, 0, sizeof(*s));
-	s->shared = TRUE;
+	s->shared = true;
 	LD->btrace_store = s;
       }
     }
@@ -149,7 +149,7 @@ next_btrace_id(btrace *bt)
 
 btrace *
 save_backtrace(const char *why)
-{ btrace *bt = get_trace_store(TRUE);
+{ btrace *bt = get_trace_store(true);
 
   if ( bt )
   { btrace_stack *s;
@@ -194,7 +194,7 @@ print_trace(btrace *bt, int me)
 
 void
 print_backtrace(int last)		/* 1..SAVE_TRACES */
-{ btrace *bt = get_trace_store(FALSE);
+{ btrace *bt = get_trace_store(false);
 
   if ( bt )
   { int me = bt->current-last;
@@ -237,9 +237,6 @@ bstore_print_backtrace_named(btrace *bt, const char *why)
 #define BTRACE_DONE 1
 #include <execinfo.h>
 #include <string.h>
-#include <dlfcn.h>
-
-#define MAXCMD 1024
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This implementation uses the libgcc unwinding capabilities. If possible,
@@ -277,7 +274,7 @@ get_trace_store(int create)
     { btrace *s = malloc(sizeof(*s));
       if ( s )
       { memset(s, 0, sizeof(*s));
-	s->shared = TRUE;
+	s->shared = true;
 	LD->btrace_store = s;
       }
     }
@@ -325,7 +322,7 @@ next_btrace_id(btrace *bt)
 
 btrace *
 save_backtrace(const char *why)
-{ btrace *bt = get_trace_store(TRUE);
+{ btrace *bt = get_trace_store(true);
 
   if ( bt )
   { void *array[100];
@@ -344,76 +341,6 @@ save_backtrace(const char *why)
   return bt;
 }
 
-#ifdef __APPLE__
-/* Emits e.g. "prologToplevel (in libswipl.8.5.20.dylib) (pl-pro.c:560)" */
-#define ADDR2LINE_CMD "atos -o \"%s\" %p"
-#else
-/* Emits two lines: "function\nfile:line"  */
-#define ADDR2LINE_CMD "addr2line -fe \"%s\" %p"
-#endif
-
-static int
-addr2line(const char *fname, uintptr_t offset, char *buf, size_t size)
-{ char cmd[MAXCMD];
-
-  if ( snprintf(cmd, size, ADDR2LINE_CMD, fname, (void*)offset) < size )
-  { FILE *fd;
-
-    if ( (fd=popen(cmd, "r")) )
-    { int c;
-      char *ebuf = &buf[size-1];
-      char *o = buf;
-      const char *sep = "() at ";
-
-#ifdef __APPLE__
-      int field = 0;
-      while((c=fgetc(fd)) != EOF && o<ebuf)
-      { if ( field == 0 )
-	{ if ( c == ' ' )
-	  { if ( o+strlen(sep) < ebuf )
-	    { strcpy(o, sep);
-	      o += strlen(sep);
-	    }
-	    field++;
-	  } else
-	  { *o++ = c;		/* copy the function */
-	  }
-	} else if ( field < 3 )
-	{ if ( c == '(' )	/* skip two '(' */
-	    field++;
-	} else
-	{ if ( c == ')' )	/* copy to ')' */
-	    break;
-	  *o++ = c;
-	}
-      }
-#else
-      int nl = 0;
-      while((c=fgetc(fd)) != EOF && o<ebuf)
-      { if ( c == '\n' )
-	{ nl++;
-
-	  if ( nl == 1 && o+strlen(sep) < ebuf)
-	  { strcpy(o, sep);
-	    o += strlen(sep);
-	  }
-	} else
-	{ *o++ = (char)c;
-	}
-      }
-#endif
-
-      *o = '\0';
-
-      pclose(fd);
-      return o > buf;
-    }
-  }
-
-  return FALSE;
-}
-
-
 static void
 print_trace(btrace *bt, int me)
 { size_t i;
@@ -422,36 +349,13 @@ print_trace(btrace *bt, int me)
   { Sdprintf("C-stack trace labeled \"%s\":\n", bt->why[me]);
 
     for(i=0; i<bt->sizes[me]; i++)
-    { Dl_info info;
+    { char buf[512];
       void *addr = bt->retaddr[me][i];
 
-#ifdef HAVE_DLADDR
-      if ( dladdr(addr, &info) )
-      { uintptr_t offset = (uintptr_t)addr - (uintptr_t)info.dli_fbase;
-
-	if ( info.dli_fname )
-	{ char buf[512];
-
-	  if ( ( strstr(info.dli_fname, ".so")
-#if __APPLE__
-		 || strstr(info.dli_fname, ".dylib")
-#endif
-	       ) &&
-	       addr2line(info.dli_fname, offset, buf, sizeof(buf)) )
-	    Sdprintf("  [%zd] %s [%p]\n", i, buf, addr);
-	  else if ( info.dli_sname )
-	    Sdprintf("  [%zd] %s(%s+0x%tx) [%p]\n",
-		     i, info.dli_fname, info.dli_sname,
-		     (char*)addr-(char*)info.dli_saddr,
-		     addr);
-	  else
-	    Sdprintf("  [%zd] %s(+%p) [%p]\n",
-		     i, info.dli_fname, (void*)offset, addr);
-	} else
-#endif
-	{ Sdprintf("  [%zd] ??? [%p]\n", i, addr);
-	}
-      }
+      if ( addr2line(addr, buf, sizeof(buf)) )
+	Sdprintf("  [%zd] %s [%p]\n", i, buf, addr);
+      else
+	Sdprintf("  [%zd] ??? [%p]\n", i, addr);
     }
   } else
   { Sdprintf("No stack trace\n");
@@ -460,7 +364,7 @@ print_trace(btrace *bt, int me)
 
 void
 print_backtrace(int last)		/* 1..SAVE_TRACES */
-{ btrace *bt = get_trace_store(FALSE);
+{ btrace *bt = get_trace_store(false);
 
   if ( bt )
   { int me = bt->current-last;
@@ -580,7 +484,7 @@ get_trace_store(int create)
   { btrace *s = malloc(sizeof(*s));
     if ( s )
     { memset(s, 0, sizeof(*s));
-      s->shared = TRUE;
+      s->shared = true;
       LD->btrace_store = s;
     }
   } else if ( create )
@@ -678,7 +582,7 @@ int backtrace(btrace_stack* trace, PEXCEPTION_POINTERS pExceptionInfo)
   ZeroMemory(&frame, sizeof( STACKFRAME64));
   memset(&moduleInfo,0,sizeof(IMAGEHLP_MODULE64));
   moduleInfo.SizeOfStruct = sizeof(IMAGEHLP_MODULE64);
-  rc = SymInitialize(hProcess, NULL, TRUE);
+  rc = SymInitialize(hProcess, NULL, true);
   if (rc == 0)
     return 0;
 
@@ -708,7 +612,7 @@ int backtrace(btrace_stack* trace, PEXCEPTION_POINTERS pExceptionInfo)
 			    SymGetModuleBase64,
 			    NULL)) != 0)
    { int hasModule = 0;
-     BOOL hasSymbol = FALSE;
+     BOOL hasSymbol = false;
 
      if (skip > 0)
      { skip--;
@@ -768,7 +672,7 @@ int backtrace(btrace_stack* trace, PEXCEPTION_POINTERS pExceptionInfo)
 
 btrace *
 win_save_backtrace(const char *why, PEXCEPTION_POINTERS pExceptionInfo)
-{ btrace *bt = get_trace_store(TRUE);
+{ btrace *bt = get_trace_store(true);
   if ( bt )
   { int current = next_btrace_id(bt);
     btrace_stack *s = &bt->dumps[current];
@@ -796,7 +700,7 @@ print_trace(btrace *bt, int me)
   { int depth;
     HANDLE hProcess = GetCurrentProcess();
 
-    SymInitialize(hProcess, NULL, TRUE);
+    SymInitialize(hProcess, NULL, true);
 
     Sdprintf("Stack trace labeled \"%s\":\n", s->name);
     for(depth=0; depth<s->depth; depth++)
@@ -840,7 +744,7 @@ print_trace(btrace *bt, int me)
 
 void
 print_backtrace(int last)		/* 1..SAVE_TRACES */
-{ btrace *bt = get_trace_store(FALSE);
+{ btrace *bt = get_trace_store(false);
 
   if ( bt )
   { int me = bt->current-last;
@@ -900,7 +804,7 @@ initBackTrace(void)
 
 void
 print_backtrace_named(const char *why)
-{ bstore_print_backtrace_named(get_trace_store(FALSE), why);
+{ bstore_print_backtrace_named(get_trace_store(false), why);
 }
 
 
@@ -915,7 +819,11 @@ print_c_backtrace(const char *why)
 
 void
 sigCrashHandler(int sig)
-{ GET_LD
+{
+#ifdef O_PLMT
+  GET_LD
+#endif
+
   signal(sig,     SIG_DFL);
 #ifdef SIGALRM
   signal(SIGALRM, SIG_DFL);
@@ -934,7 +842,7 @@ sigCrashHandler(int sig)
 	   sig, signal_name(sig));
   save_backtrace("crash");
 #ifdef O_PLMT
-  LD->thread.info->c_stack_low = TRUE; /* Actually, we are on an alt-stack */
+  LD->thread.info->c_stack_low = true; /* Actually, we are on an alt-stack */
 #endif
   printCrashContext("crash");
 
@@ -984,7 +892,7 @@ PRED_IMPL("c_backtrace_clear", 0, c_backtrace_clear, 0)
     LD->btrace_store = NULL;
   }
 
-  return TRUE;
+  return true;
 }
 
 static
@@ -993,10 +901,10 @@ PRED_IMPL("c_backtrace_print", 1, c_backtrace_print, 0)
 
   if ( PL_get_chars(A1, &s, CVT_ATOM|CVT_STRING|CVT_EXCEPTION) )
   { print_backtrace_named(s);
-    return TRUE;
+    return true;
   }
 
-  return FALSE;
+  return false;
 }
 
 #endif
@@ -1182,7 +1090,7 @@ CStackSize(DECL_LD)
 #endif
       }
     }
-    cinfo->initialised = TRUE;
+    cinfo->initialised = true;
     DEBUG(MSG_CSTACK, Sdprintf("[%d]: C-stack %p[%zd]\n",
 			       info->pl_tid, cinfo->base, cinfo->size));
 
@@ -1192,4 +1100,113 @@ CStackSize(DECL_LD)
 #else /*O_PLMT*/
   return NULL;
 #endif
+}
+
+		/*******************************
+		*          ADDR2LINE           *
+		*******************************/
+
+#if !defined(__WINDOWS__) && !defined(__EMSCRIPTEN__)
+#include <stdio.h>
+#include <dlfcn.h>
+#define HAVE_ADDR2LINE2 1
+#define MAXCMD 1024
+
+#ifdef __APPLE__
+/* Emits e.g. "prologToplevel (in libswipl.8.5.20.dylib) (pl-pro.c:560)" */
+#define ADDR2LINE_CMD "atos -o \"%s\" --fullPath %p"
+#else
+/* Emits two lines: "function\nfile:line"  */
+#define ADDR2LINE_CMD "addr2line -fe \"%s\" %p"
+#endif
+
+static int
+addr2line2(const char *fname, uintptr_t offset, char *buf, size_t size)
+{ char cmd[MAXCMD];
+
+  if ( snprintf(cmd, size, ADDR2LINE_CMD, fname, (void*)offset) < size )
+  { FILE *fd;
+
+    if ( (fd=popen(cmd, "r")) )
+    { int c;
+      char *ebuf = &buf[size-1];
+      char *o = buf;
+      const char *sep = "() at ";
+
+#ifdef __APPLE__
+      int field = 0;
+      while((c=fgetc(fd)) != EOF && o<ebuf)
+      { if ( field == 0 )
+	{ if ( c == ' ' )
+	  { if ( o+strlen(sep) < ebuf )
+	    { strcpy(o, sep);
+	      o += strlen(sep);
+	    }
+	    field++;
+	  } else
+	  { *o++ = c;		/* copy the function */
+	  }
+	} else if ( field < 3 )
+	{ if ( c == '(' )	/* skip two '(' */
+	    field++;
+	} else
+	{ if ( c == ')' )	/* copy to ')' */
+	    break;
+	  *o++ = c;
+	}
+      }
+#else
+      int nl = 0;
+      while((c=fgetc(fd)) != EOF && o<ebuf)
+      { if ( c == '\n' )
+	{ nl++;
+
+	  if ( nl == 1 && o+strlen(sep) < ebuf)
+	  { strcpy(o, sep);
+	    o += strlen(sep);
+	  }
+	} else
+	{ *o++ = (char)c;
+	}
+      }
+#endif
+
+      *o = '\0';
+
+      pclose(fd);
+      return o > buf;
+    }
+  }
+
+  return false;
+}
+#endif/*!__WINDOWS__ && !__EMSCRIPTEN__ */
+
+bool
+addr2line(void *addr, char *buf, size_t size)
+{
+#if HAVE_DLADDR && HAVE_ADDR2LINE2
+  Dl_info info;
+
+  if ( dladdr(addr, &info) )
+  { uintptr_t offset = (uintptr_t)addr - (uintptr_t)info.dli_fbase;
+
+    if ( info.dli_fname )
+    { if ( ( strstr(info.dli_fname, ".so")
+#if __APPLE__
+	     || strstr(info.dli_fname, ".dylib")
+#endif
+	   ) )
+      { return addr2line2(info.dli_fname, offset, buf, size);
+      } else if ( info.dli_sname )
+      { snprintf(buf, size, "%s(%s+0x%tx)",
+		 info.dli_fname, info.dli_sname,
+		 (char*)addr-(char*)info.dli_saddr);
+	return true;
+      }
+    }
+  }
+#endif/*HAVE_DLADDR*/
+
+  return false;
 }
