@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2024, University of Amsterdam
+    Copyright (c)  2011-2025, University of Amsterdam
 			      VU University Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -51,17 +51,13 @@ source should also use format() to produce error messages, etc.
 #include <stdio.h>
 #include <math.h>
 #include <fenv.h>
+#include "pl-fmt.h"
 #ifdef __WINDOWS__
 #include "../pl-nt.h"
 #endif
 
 typedef foreign_t (*Func1)(term_t a1);
 
-static char *	formatInteger(PL_locale *locale, int div, int radix,
-			     bool smll, Number n, Buffer out);
-static char *	formatFloat(PL_locale *locale, int how, int arg,
-			    Number f, Buffer out);
-
 #define MAXRUBBER 100
 
 struct rubber
@@ -73,22 +69,29 @@ struct rubber
 typedef struct
 { IOSTREAM *out;			/* our output stream */
   int column;				/* current column */
+  int tab_stop;				/* Last tab stop */
   tmp_buffer buffer;			/* bin for characters with tabs */
   size_t buffered;			/* characters in buffer */
   int pending_rubber;			/* number of not-filled ~t's */
   struct rubber rub[MAXRUBBER];
 } format_state;
 
+static char *	formatFloat(PL_locale *locale, int how, int arg,
+			    Number f, Buffer out);
+static void	distribute_rubber(struct rubber *, int, int);
+static WUNUSED int emit_rubber(format_state *state);
+
 #define BUFSIZE		1024
-#define DEFAULT		(-1)
+#define DEFAULT		INT_MIN
 #define SHIFT		{ argc--; argv++; }
 #define NEED_ARG	{ if ( argc <= 0 ) \
 			  { FMT_ERROR("not enough arguments"); \
 			  } \
 			}
 #define FMT_ERROR(fmt)	return PL_error(NULL, 0, NULL, ERR_FORMAT, fmt)
-#define FMT_ARG(c, a)	return PL_error(NULL, 0, NULL, ERR_FORMAT_ARG, c, a)
-#define FMT_EXEPTION()	return FALSE
+#define FMT_ARG(s, a)	return PL_error(NULL, 0, NULL, ERR_FORMAT_ARG, s, a)
+#define FMT_ARGC(c, a)  do { char f[2] = {(char)c}; FMT_ARG(f, a); } while(0)
+#define FMT_EXEPTION()	return false
 
 
 static PL_locale prolog_locale =
@@ -97,14 +100,25 @@ static PL_locale prolog_locale =
 };
 
 
-static int
-update_column(int col, int c)
-{ switch(c)
-  { case '\n':	return 0;
-    case '\r':  return 0;
-    case '\t':	return (col + 1) | 0x7;
-    case '\b':	return (col <= 0 ? 0 : col - 1);
-    default:	return col + 1;
+static inline void
+update_column(format_state *state, int c)
+{ if ( likely(c >= ' ') )
+  { state->column++;
+  } else
+  { switch(c)
+    { case '\n':
+	state->column = 0;
+	state->tab_stop = 0;
+	break;
+      case '\t':
+	state->column = (state->column+1)|0x7;
+      case '\b':
+	if ( likely(state->column>0) )
+	  state->column--;
+	break;
+      default:
+	state->column++;
+    }
   }
 }
 
@@ -115,7 +129,7 @@ UTF-8 format in the state's `buffer'.   The  `buffered' field represents
 the number of UTF-8 characters in the buffer.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static WUNUSED int
+static WUNUSED bool
 outchr(format_state *state, int chr)
 { if ( state->pending_rubber )
   { if ( chr > 0x7f )
@@ -132,12 +146,12 @@ outchr(format_state *state, int chr)
     state->buffered++;
   } else
   { if ( Sputcode(chr, state->out) < 0 )
-      return FALSE;
+      return false;
   }
 
-  state->column = update_column(state->column, chr);
+  update_column(state, chr);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -157,40 +171,55 @@ outstring(format_state *state, const char *s, size_t len)
   } else
   { for(q=s; q < e; q++)
     { if ( Sputcode(*q&0xff, state->out) < 0 )
-	return FALSE;
+	return false;
     }
   }
 
   for(q=s; q < e; q++)
-    state->column = update_column(state->column, *q&0xff);
+    update_column(state, *q&0xff);
 
-  return TRUE;
+  return true;
 }
 
 
-static WUNUSED int
+static WUNUSED bool
 oututf8(format_state *state, const char *s, size_t len)
 { const char *e = &s[len];
 
-  while(s<e)
-  { int chr;
+  if ( !state->pending_rubber )
+  { while(s<e)
+    { int chr = s[0]&0xff;
 
-    PL_utf8_code_point(&s, e, &chr);
-    if ( !outchr(state, chr) )
-      return FALSE;
+      if ( likely(chr < 0x80) )
+	s++;
+      else
+	PL_utf8_code_point(&s, e, &chr);
+
+      if ( Sputcode(chr, state->out) < 0 )
+	return false;
+      update_column(state, chr);
+    }
+  } else
+  { while(s<e)
+    { int chr;
+
+      PL_utf8_code_point(&s, e, &chr);
+      if ( !outchr(state, chr) )
+	return false;
+    }
   }
 
-  return TRUE;
+  return true;
 }
 
 
-static WUNUSED int
+static WUNUSED bool
 oututf80(format_state *state, const char *s)
 { return oututf8(state, s, strlen(s));
 }
 
 
-static WUNUSED int
+static WUNUSED bool
 outtext(format_state *state, PL_chars_t *txt)
 { switch(txt->encoding)
   { case ENC_ISO_LATIN_1:
@@ -204,14 +233,14 @@ outtext(format_state *state, PL_chars_t *txt)
 
 	s = get_wchar(s, &c);
 	if ( !outchr(state, c) )
-	  return FALSE;
+	  return false;
       }
 
-      return TRUE;
+      return true;
     }
     default:
     { assert(0);
-      return FALSE;
+      return false;
     }
   }
 }
@@ -219,13 +248,6 @@ outtext(format_state *state, PL_chars_t *txt)
 
 #define format_predicates (GD->format.predicates)
 
-static int	update_column(int, Char);
-static bool	do_format(IOSTREAM *fd, PL_chars_t *fmt,
-			  int ac, term_t av, Module m);
-static void	distribute_rubber(struct rubber *, int, int);
-static WUNUSED int emit_rubber(format_state *state);
-
-
 		/********************************
 		*       PROLOG CONNECTION	*
 		********************************/
@@ -237,10 +259,10 @@ PRED_IMPL("format_predicate", 2, format_predicate, META)
   predicate_t proc = NULL;
   size_t arity;
 
-  if ( !PL_get_char_ex(A1, &c, FALSE) )
-    return FALSE;
+  if ( !PL_get_char_ex(A1, &c, false) )
+    return false;
   if ( !get_procedure(A2, &proc, 0, GP_CREATE) )
-    return FALSE;
+    return false;
 
   PL_predicate_info(proc, NULL, &arity, NULL);
   if ( arity == 0 )
@@ -253,7 +275,7 @@ PRED_IMPL("format_predicate", 2, format_predicate, META)
 
   updateHTableWP(format_predicates, c, proc);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -279,12 +301,12 @@ PRED_IMPL("current_format_predicate", 2, current_format_predicate, NDET|META)
       e = CTX_PTR;
       freeTableEnum(e);
     default:
-      return TRUE;
+      return true;
   }
 
   if ( !(fid = PL_open_foreign_frame()) )
   { freeTableEnum(e);
-    return FALSE;
+    return false;
   }
 
   table_key_t tk;
@@ -304,7 +326,7 @@ PRED_IMPL("current_format_predicate", 2, current_format_predicate, NDET|META)
 
   PL_close_foreign_frame(fid);
   freeTableEnum(e);
-  return FALSE;
+  return false;
 }
 
 
@@ -320,23 +342,31 @@ format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
   if ( !PL_get_text(format, &fmt, CVT_ATOM|CVT_STRING|CVT_LIST|BUF_STACK) )
     return PL_error("format", 3, NULL, ERR_TYPE, ATOM_text, format);
 
-  if ( (argc = (int)lengthList(args, FALSE)) >= 0 )
+  intptr_t len = lengthList(args, false);
+  if ( len >= 0 )
   { term_t head = PL_new_term_ref();
     int n = 0;
 
+    argc = len;
     argv = PL_new_term_refs(argc);
     while( PL_get_list(args, head, args) )
       PL_put_term(argv+n++, head);
   } else
-  { argc = 1;
-    argv = PL_new_term_refs(argc);
+  { Word p = valTermRef(args);
+    deRef(p);
 
-    PL_put_term(argv, args);
+    if ( !isList(*p) )
+    { argc = 1;
+      argv = PL_new_term_refs(argc);
+
+      PL_put_term(argv, args);
+    } else
+      return PL_type_error("list", args);
   }
 
   switch(fmt.storage)			/* format can do call-back! */
-  { case PL_CHARS_RING:
-    case PL_CHARS_STACK:
+  { case PL_CHARS_STACK:
+    case PL_CHARS_PROLOG_STACK:
       PL_save_text(&fmt, BUF_MALLOC);
       break;
     default:
@@ -361,9 +391,9 @@ format(DECL_LD term_t out, term_t format, term_t args)
   term_t list = PL_new_term_ref();
 
   if ( !PL_strip_module(args, &m, list) )
-    return FALSE;
+    return false;
 
-  if ( (rc=setupOutputRedirect(out, &ctx, FALSE)) )
+  if ( (rc=setupOutputRedirect(out, &ctx, false)) )
   { if ( (rc = format_impl(ctx.stream, format, list, m)) )
       rc = closeOutputRedirect(&ctx);
     else
@@ -384,7 +414,7 @@ static
 PRED_IMPL("format", 3, format, META)
 { PRED_LD
 
-    return format(A1, A2, A3);
+  return format(A1, A2, A3);
 }
 
 static inline int
@@ -425,11 +455,11 @@ prepare_sub_format(sub_state *state, format_state *fstate, IOSTREAM *fd)
     state->bufsize    = sizeof(state->buf);
 
     tellString(&state->str, &state->bufsize, ENC_UTF8);
-    if ( true(fd, SIO_ISATTY) )
+    if ( ison(fd, SIO_ISATTY) )
       set(Scurout, SIO_ISATTY);
   }
 
-  return TRUE;
+  return true;
 }
 
 static int
@@ -455,23 +485,41 @@ end_sub_format(sub_state *state, int rc)
   return rc;
 }
 
+static bool
+get_int_expression(term_t t, Number i)
+{ if ( !PL_get_number(t, i) &&
+       !valueExpression(t, i) )
+    return false;
 
+  if ( !isIntegerNumber(i) &&
+       !toIntegerNumber(i, 0) )
+    return false;
 
+  if ( i->type == V_MPZ )
+  { int64_t v;
+    if ( mpz_to_int64(i->value.mpz, &v) )
+    { clearNumber(i);
+      i->type = V_INTEGER;
+      i->value.i = v;
+    }
+  }
 
+  return true;
+}
 
 		/********************************
 		*       ACTUAL FORMATTING	*
 		********************************/
 
-static bool
+bool
 do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 { GET_LD
   format_state state;			/* complete state */
-  int tab_stop = 0;			/* padded tab stop */
   unsigned int here = 0;
-  int rc = TRUE;
+  int rc = true;
 
   state.out = fd;
+  state.tab_stop = 0;
   state.pending_rubber = 0;
   initBuffer(&state.buffer);
   state.buffered = 0;
@@ -487,10 +535,18 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
     switch(c)
     { case '~':
 	{ int arg = DEFAULT;		/* Numeric argument */
-	  int mod_colon = FALSE;	/* Used colon modifier */
+	  int mod_colon = false;	/* Used colon modifier */
+	  bool neg = false;
 	  predicate_t proc;
 					/* Get the numeric argument */
 	  c = get_chr_from_text(fmt, ++here);
+	  if ( c == '-' )
+	  { neg = true;
+	    c = get_chr_from_text(fmt, ++here);
+	    if ( !isDigitW(c) )
+	    { FMT_ERROR("invalid argument");
+	    }
+	  }
 
 	  if ( isDigitW(c) )
 	  { arg = c - '0';
@@ -509,14 +565,27 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		arg = arg2;
 		here++;
 	      } else
+	      { if ( neg )
+		  arg = -arg;
 		break;
+	      }
 	    }
 	  } else if ( c == '*' )
-	  { NEED_ARG;
-	    if ( PL_get_integer(argv, &arg) && arg >= 0 )
-	    { SHIFT;
-	    } else
-	      FMT_ERROR("no or negative integer for `*' argument");
+	  { number i;
+	    NEED_ARG;
+	    AR_CTX
+
+	    AR_BEGIN();
+	    if ( !get_int_expression(argv, &i) ||
+		 i.type != V_INTEGER ||
+		 i.value.i < 0 ||
+		 i.value.i > INT_MAX )
+	    { AR_CLEANUP();
+	      FMT_ARGC(c, argv);
+	    }
+	    SHIFT;
+	    arg = i.value.i;
+	    AR_END();
 	    c = get_chr_from_text(fmt, ++here);
 	  } else if ( c == '`' && here < fmt->length )
 	  { arg = get_chr_from_text(fmt, ++here);
@@ -524,7 +593,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	  }
 
 	  if ( c == ':' )
-	  { mod_colon = TRUE;
+	  { mod_colon = true;
 	    c = get_chr_from_text(fmt, ++here);
 	  }
 
@@ -565,7 +634,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		{ PL_chars_t txt;
 
 		  NEED_ARG;
-		  if ( !PL_get_text(argv, &txt, CVT_ATOMIC) )
+		  if ( !PL_get_text(argv, &txt, CVT_ATOM|CVT_STRING) )
 		    FMT_ARG("a", argv);
 		  SHIFT;
 		  rc = outtext(&state, &txt);
@@ -595,29 +664,34 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      case 'e':			/* exponential float */
 	      case 'E':			/* Exponential float */
 	      case 'f':			/* float */
+	      case 'F':			/* float */
 	      case 'g':			/* shortest of 'f' and 'e' */
 	      case 'G':			/* shortest of 'f' and 'E' */
-		{ AR_CTX
-		  number n;
-		  union {
-		  tmp_buffer b;
+	      case 'h':
+	      case 'H':			/* Precise */
+		{ number n;
+		  union
+		  { tmp_buffer b;
 		    buffer b1;
 		  } u;
 		  PL_locale *l;
+		  AR_CTX
 
 		  NEED_ARG;
 		  AR_BEGIN();
-		  if ( !valueExpression(argv, &n) )
-		  { char f[2];
+		  if ( !PL_get_number(argv, &n) )
+		  { if ( !valueExpression(argv, &n) )
+		    { char f[2];
 
-		    f[0] = (char)c;
-		    f[1] = EOS;
-		    AR_CLEANUP();
-		    FMT_ARG(f, argv); /* returns error */
+		      f[0] = (char)c;
+		      f[1] = EOS;
+		      AR_CLEANUP();
+		      FMT_ARG(f, argv); /* returns error */
+		    }
 		  }
 		  SHIFT;
 
-		  if ( c == 'f' && mod_colon )
+		  if ( (c == 'f' || c == 'F') && mod_colon )
 		    l = fd->locale;
 		  else
 		    l = &prolog_locale;
@@ -639,23 +713,19 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      case 'r':			/* radix number */
 	      case 'R':			/* Radix number */
 	      case 'I':			/* Prolog 1_000_000 */
-		{ AR_CTX
-		  number i;
+		{ number i;
 		  tmp_buffer b;
 		  char *si;
+		  AR_CTX
 
 		  NEED_ARG;
 		  AR_BEGIN();
-		  if ( !valueExpression(argv, &i) ||
-		       !toIntegerNumber(&i, 0) )
-		  { char f[2];
-
-		    f[0] = (char)c;
-		    f[1] = EOS;
-		    AR_CLEANUP();
-		    FMT_ARG(f, argv);
+		  if ( !get_int_expression(argv, &i) )
+		  { AR_CLEANUP();
+		    FMT_ARGC(c, argv); /* return with error */
 		  }
 		  SHIFT;
+
 		  initBuffer(&b);
 		  if ( c == 'd' || c == 'D' )
 		  { PL_locale ltmp;
@@ -663,9 +733,15 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		    static char grouping[] = {3,0};
 
 		    if ( c == 'D' )
-		    { ltmp.thousands_sep = L",";
-		      ltmp.decimal_point = L".";
-		      ltmp.grouping = grouping;
+		    { if ( mod_colon )
+		      { ltmp.thousands_sep = L"_";
+			ltmp.decimal_point = L".";
+			ltmp.grouping = grouping;
+		      } else
+		      { ltmp.thousands_sep = L",";
+			ltmp.decimal_point = L".";
+			ltmp.grouping = grouping;
+		      }
 		      l = &ltmp;
 		    } else if ( mod_colon )
 		    { l = fd->locale;
@@ -675,7 +751,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 
 		    if ( arg == DEFAULT )
 		      arg = 0;
-		    si = formatInteger(l, arg, 10, TRUE, &i, (Buffer)&b);
+		    si = formatInteger(l, arg, 10, true, &i, (Buffer)&b);
 		  } else if ( c == 'I' )
 		  { PL_locale ltmp;
 		    char grouping[2];
@@ -685,7 +761,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		    ltmp.thousands_sep = L"_";
 		    ltmp.grouping = grouping;
 
-		    si = formatInteger(&ltmp, 0, 10, TRUE, &i, (Buffer)&b);
+		    si = formatInteger(&ltmp, 0, 10, true, &i, (Buffer)&b);
 		  } else			/* r,R */
 		  { if ( arg == DEFAULT )
 		      arg = 8;
@@ -703,7 +779,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  if ( si )
 		    rc = oututf80(&state, si);
 		  else
-		    rc = FALSE;
+		    rc = false;
 		  discardBuffer(&b);
 		  if ( !rc )
 		    goto out;
@@ -716,13 +792,17 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  NEED_ARG;
 		  if ( !PL_get_text(argv, &txt, CVT_ATOM|CVT_LIST|CVT_STRING) )
 		    FMT_ARG("s", argv);
-		  if ( arg != DEFAULT )
-		  { if ( arg < 0 )
-		      arg = 0;
-		    if ( arg < txt.length )
-		      txt.length = arg;
-		  }
+
+		  if ( arg == DEFAULT )
+		    arg = txt.length;
+		  else if ( arg < 0 )
+		    arg = 0;
+		  else if ( arg < txt.length )
+		    txt.length = arg;
+
 		  rc = outtext(&state, &txt);
+		  for(int i=arg-txt.length; rc && i>0; i--)
+		    rc = outchr(&state, ' ');
 		  PL_free_text(&txt);
 		  SHIFT;
 		  if ( !rc )
@@ -822,11 +902,8 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  break;
 		}
 	      case 'n':			/* \n */
-	      case 'N':			/* \n if not on newline */
 		{ if ( arg == DEFAULT )
 		    arg = 1;
-		  if ( c == 'N' && state.column == 0 )
-		    arg--;
 		  while( arg-- > 0 )
 		  { rc = outchr(&state, '\n');
 		    if ( !rc )
@@ -835,6 +912,14 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  here++;
 		  break;
 		}
+	      case 'N':			/* \n if not on newline */
+		if ( state.column != 0 )
+		{ rc = outchr(&state, '\n');
+		  if ( !rc )
+		    goto out;
+		}
+		here++;
+		break;
 	      case 't':			/* insert tab */
 		{ if ( state.pending_rubber >= MAXRUBBER )
 		    FMT_ERROR("Too many tab stops");
@@ -858,7 +943,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      case '+':			/* tab relative */
 		  if ( arg == DEFAULT )
 		    arg = 8;
-		  stop = (c == '+' ? tab_stop + arg : arg);
+		  stop = (c == '+' ? state.tab_stop + arg : arg);
 
 		  if ( stop < state.column && mod_colon )
 		    nl_and_reindent = state.pending_rubber ?
@@ -879,10 +964,10 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 
 		  if ( nl_and_reindent )
 		  { if ( Sputcode('\n', state.out) < 0 )
-		    { rc = FALSE;
+		    { rc = false;
 		      goto out;
 		    }
-		    state.column = update_column(state.column, '\n');
+		    update_column(&state, '\n');
 
 		    state.rub[0].where = state.buffered;
 		    state.rub[0].pad = nl_and_reindent;
@@ -895,7 +980,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		      goto out;
 		  }
 
-		  state.column = tab_stop = stop;
+		  state.column = state.tab_stop = stop;
 		  here++;
 		  break;
 		}
@@ -937,15 +1022,13 @@ static void
 distribute_rubber(struct rubber *r, int rn, int space)
 { if ( space > 0 )
   { int s = space / rn;
-    int n, m;
 
-    for(n=0; n < rn; n++)		/* give them equal size */
+    for(int n=0; n < rn; n++)		/* give them equal size */
       r[n].size = s;
 					/* distribute from the center */
     space -= s*rn;
-    for(m = rn / 2, n = 0; space; n++, space--)
-    { r[m + (n % 2 ? n : -n)].size++;
-    }
+    for(int n=rn-1; space; n--, space--)
+      r[n].size++;
   } else
   { int n;
 
@@ -971,7 +1054,7 @@ emit_rubber(format_state *state)
 
       for(n=0; n<r->size; n++)
       { if ( Sputcode(r->pad, state->out) < 0 )
-	  return FALSE;
+	  return false;
       }
       r++;
       rn--;
@@ -980,7 +1063,7 @@ emit_rubber(format_state *state)
     if ( s < e )
     { PL_utf8_code_point(&s, e, &chr);
       if ( Sputcode(chr, state->out) < 0 )
-	return FALSE;
+	return false;
     } else
       break;
   }
@@ -990,7 +1073,7 @@ emit_rubber(format_state *state)
   state->buffered = 0;
   state->pending_rubber = 0;
 
-  return TRUE;
+  return true;
 }
 
 
@@ -1039,7 +1122,15 @@ revert_string(char *s, size_t len)
   }
 }
 
-static char *
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Format the integer  `i` to the buffer `out`.  `div`  is for supporting
+fixed point numbers.   `radix` is the base and  `smll` defines whether
+to use capitals (`false`) or  lowercase letters for digit values above
+9.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+
+char *
 formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
 	     Buffer out)
 { const char *grouping = NULL;
@@ -1059,8 +1150,8 @@ formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
       if ( n == 0 && div == 0 )
       { addBuffer(out, '0', char);
       } else
-      { int before = FALSE;			/* before decimal point */
-	int negative = FALSE;
+      { int before = false;			/* before decimal point */
+	int negative = false;
 	int gsize = 0;
 	int dweight;
 
@@ -1070,7 +1161,7 @@ formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
 	{ if ( div-- == 0 && !before )
 	  { if ( !isEmptyBuffer(out) )
 	      lappend(locale->decimal_point, '.', out);
-	    before = TRUE;
+	    before = true;
 	    if ( grouping )
 	      gsize = grouping[0];
 	  }
@@ -1117,9 +1208,9 @@ formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
 	buf = tmp;
 
       EXCEPTION_GUARDED({ mpz_get_str(buf, radix, i->value.mpz);
-			  rc = TRUE;
+			  rc = true;
 			},
-			{ rc = FALSE;
+			{ rc = false;
 			});
       if ( !rc )
 	return NULL;
@@ -1132,7 +1223,7 @@ formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
       }
 
       if ( grouping || div > 0 )
-      { int before = FALSE;			/* before decimal point */
+      { int before = false;			/* before decimal point */
 	int gsize = 0;
 	char *e = buf+strlen(buf)-1;
 
@@ -1140,7 +1231,7 @@ formatInteger(PL_locale *locale, int div, int radix, bool smll, Number i,
 	{ if ( div-- == 0 && !before )
 	  { if ( !isEmptyBuffer(out) )
 	      lappend(locale->decimal_point, '.', out);
-	    before = TRUE;
+	    before = true;
 	    if ( grouping )
 	      gsize = grouping[0];
 	  }
@@ -1219,11 +1310,11 @@ static int
 same_decimal_point(PL_locale *l1, PL_locale *l2)
 { if ( l1->decimal_point && l2->decimal_point &&
        wcscmp(l1->decimal_point, l2->decimal_point) == 0 )
-    return TRUE;
+    return true;
   if ( !l1->decimal_point && !l2->decimal_point )
-    return TRUE;
+    return true;
 
-  return FALSE;
+  return false;
 }
 
 
@@ -1231,7 +1322,7 @@ static int
 utf8_dp(PL_locale *l, char *s, int *len)
 { if ( l->decimal_point )
   { if ( !ths_to_utf8(s, l->decimal_point, 20) )
-      return FALSE;
+      return false;
     *len = (int)strlen(s);
   } else
   { *s++ = '.';
@@ -1239,7 +1330,7 @@ utf8_dp(PL_locale *l, char *s, int *len)
     *len = 1;
   }
 
-  return TRUE;
+  return true;
 }
 
 
@@ -1253,7 +1344,7 @@ static int
 localizeDecimalPoint(PL_locale *locale, Buffer b)
 { if ( locale == GD->locale.default_locale ||
        same_decimal_point(GD->locale.default_locale, locale) )
-    return TRUE;
+    return true;
 
   if ( locale->decimal_point && locale->decimal_point[0] )
   { char *s = baseBuffer(b, char);
@@ -1263,7 +1354,7 @@ localizeDecimalPoint(PL_locale *locale, Buffer b)
 
     if ( !utf8_dp(locale, dp, &dplen) ||
 	 !utf8_dp(GD->locale.default_locale, ddp, &ddplen) )
-      return FALSE;
+      return false;
 
     if ( *s == '-' )
       s++;
@@ -1285,7 +1376,7 @@ localizeDecimalPoint(PL_locale *locale, Buffer b)
     }
   }
 
-  return TRUE;
+  return true;
 }
 
 
@@ -1312,7 +1403,7 @@ groupDigits(PL_locale *locale, Buffer b)
       int thslen;
 
       if ( !ths_to_utf8(ths, locale->thousands_sep, sizeof(ths)) )
-	return FALSE;
+	return false;
       thslen = (int)strlen(ths);
 
       if ( !growBuffer(b, thslen*groups) )
@@ -1337,7 +1428,7 @@ groupDigits(PL_locale *locale, Buffer b)
     }
   }
 
-  return TRUE;
+  return true;
 }
 
 
@@ -1345,7 +1436,7 @@ groupDigits(PL_locale *locale, Buffer b)
 formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
 
 formats a floating point  number  to  a   buffer.  `How'  is  the format
-specifier ([eEfgG]), `arg' the argument.
+specifier ([eEfFgG]), `arg' the argument.
 
 MPZ/MPQ numbers printed using the format specifier `f' are written using
 the following algorithm, courtesy of Jan Burse:
@@ -1358,9 +1449,11 @@ the following algorithm, courtesy of Jan Burse:
 
 static char *
 formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
-{ if ( arg == DEFAULT )
-  { arg = 6;
-  } else if ( arg < 0 )
+{ bool use_h = how == 'h' || how == 'H';
+
+  if ( arg == DEFAULT )
+  { arg = use_h ? 3 : 6;
+  } else if ( arg < 0 && !use_h )
   { GET_LD
     term_t a = PL_new_term_ref();
     PL_put_integer(a, arg);
@@ -1368,6 +1461,9 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
     return NULL;
   }
 
+  if ( use_h && !promoteToFloatNumber(f) )
+    return NULL;
+
   switch(f->type)
   {
 #ifdef O_GMP
@@ -1379,6 +1475,7 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
     case V_MPZ:
     { switch(how)
       { case 'f':
+	case 'F':
 	{ mpz_init(t1);
 	  mpz_init(t2);
 	  mpz_ui_pow_ui(t1, 10, arg);
@@ -1434,6 +1531,7 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
 
       switch(how)
       { case 'f':
+	case 'F':
 	{ mpz_init(t1);
 	  mpz_init(t2);
 	  mpz_ui_pow_ui(t1, 10, arg);
@@ -1649,7 +1747,7 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
       int upcase;
       limb_t prec = ((double)(arg+2) * log(10)/log(2));
 
-      if ( how == 'f' )		/* we must compensate for the integer */
+      if ( how == 'f' || how == 'F' ) /* we must compensate for the integer */
       { mpz_t i;
 	mpz_init(i);
 	mpz_set_q(i, f->value.mpq);
@@ -1659,19 +1757,20 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
       mpz_init(n);
       bf_div(n, mpq_numref(f->value.mpq), mpq_denref(f->value.mpq), prec, BF_RNDN);
     bf_print:
-      upcase = FALSE;
+      upcase = false;
       switch(how)
       { case 'f':
+	case 'F':
 	  flags = BF_FTOA_FORMAT_FRAC;
 	  break;
 	case 'E':
-	  upcase = TRUE;
+	  upcase = true;
 	case 'e':
 	  arg++;		/* LibBF counts total, we after . */
 	  flags = BF_FTOA_FORCE_EXP|BF_FTOA_FORMAT_FIXED;
 	  break;
 	case 'G':
-	  upcase = TRUE;
+	  upcase = true;
 	case 'g':
 	  flags = BF_FTOA_FORMAT_FREE;
 	  break;
@@ -1698,26 +1797,48 @@ formatFloat(PL_locale *locale, int how, int arg, Number f, Buffer out)
   }
 #endif /* O_GMP OR O_BF */
     case V_INTEGER:
-      promoteToFloatNumber(f);
+      if ( !promoteToFloatNumber(f) )
+	return NULL;
       /*FALLTHROUGH*/
     case V_FLOAT:
     { char tmp[12];
       int written = arg+20;
       int size = 0;
 
-      Ssprintf(tmp, "%%.%d%c", arg, how);
-      while(written >= size)
-      { size = written+1;
+      if ( how == 'h' || how == 'H' )
+      { size_t space = 32;
 
-	if ( !growBuffer(out, size) )
-	{ PL_no_memory();
-	  return NULL;
+	for(int n=0; n<2; n++)
+	{ size_t sz;
+
+	  if ( !growBuffer(out, space) )
+	  { PL_no_memory();
+	    return NULL;
+	  }
+	  sz = format_float(out->base, space, f->value.f,
+			    arg, how == 'H' ? 'E' : 'e');
+	  if ( sz < space )
+	  { written = sz;
+	    break;
+	  } else
+	  { space = sz+1;
+	  }
 	}
-	written = snprintf(baseBuffer(out, char), size, tmp, f->value.f);
+      } else
+      { Ssprintf(tmp, "%%.%d%c", arg, how);
+	while(written >= size)
+	{ size = written+1;
+
+	  if ( !growBuffer(out, size) )
+	  { PL_no_memory();
+	    return NULL;
+	  }
+	  written = snprintf(baseBuffer(out, char), size, tmp, f->value.f);
 #ifdef __WINDOWS__
-	if ( written < 0 )	/* pre-C99 Windows snprintf() returns -1 */
-	  written = size*2;
+	  if ( written < 0 )	/* pre-C99 Windows snprintf() returns -1 */
+	    written = size*2;
 #endif
+	}
       }
 
 #ifdef __WINDOWS__
