@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2024, University of Amsterdam
+    Copyright (c)  2011-2025, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
@@ -36,10 +36,12 @@
 */
 
 #ifdef __WINDOWS__
+#define SWIPL_WINDOWS_NATIVE_ACCESS 1
 #include "windows/uxnt.h"
 #include "config/wincfg.h"
 #include <winsock2.h>
 #include "../pl-nt.h"
+#include "../pl-ntconsole.h"
 #define CRLF_MAPPING 1
 #else
 #include <config.h>
@@ -874,14 +876,14 @@ reperror(int c, IOSTREAM *s)
     const char *q;
 
     if ( (s->flags & SIO_REPPL) )
-    { sprintf(buf, "\\x%X\\", c);
+    { snprintf(buf, sizeof buf, "\\x%X\\", c);
     } else if ( (s->flags & SIO_REPPLU) )
     { if ( c <= 0xffff )
-	sprintf(buf, "\\u%04X", c);
+	snprintf(buf, sizeof buf, "\\u%04X", c);
       else
-	sprintf(buf, "\\U%08X", c);
+	snprintf(buf, sizeof buf, "\\U%08X", c);
     } else
-    { sprintf(buf, "&#%d;", c);
+    { snprintf(buf, sizeof buf, "&#%d;", c);
     }
 
     for(q = buf; *q; q++)
@@ -1725,6 +1727,15 @@ Sset_exception(IOSTREAM *s, term_t ex)
       } else
       { r = s->exception = PL_record(ex);
       }
+
+      /* If the current exception is associated with the
+       * stream we should clear it.  It will be re-raised
+       * by reportStreamError(), which clears the exception
+       * from the stream again.
+       */
+      term_t pending = PL_exception(0);
+      if ( pending && PL_compare(ex,pending) == CMP_EQUAL )
+	PL_clear_exception();
     }
 
     s->flags = nflags;
@@ -2006,6 +2017,37 @@ Stell(IOSTREAM *s)
 }
 
 
+/**
+ * Set/get the notion  of the size of the console.   Typically used on
+ * `Suser_output`  on systems  where we  have  no other  means to  get
+ * access to the size.  This  notably concerns Epilog on Windows which
+ * communicates using  Windows pipes.  This  implies we have  no POSIX
+ * tty/pty, not a Windows (pseudo) console.
+ */
+
+int
+Ssetttysize(IOSTREAM *s, short cols, short rows)
+{ if ( s->magic != SIO_MAGIC )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  s->tty_size = (cols<<16) + rows;
+  return 0;
+}
+
+int
+Sgetttysize(IOSTREAM *s, short *cols, short *rows)
+{ if ( s->magic != SIO_MAGIC )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  *cols = s->tty_size >> 16;
+  *rows = s->tty_size & 0xffff;
+  return 0;
+}
+
 		 /*******************************
 		 *	      CLOSE		*
 		 *******************************/
@@ -3304,16 +3346,6 @@ For now, we use PL_malloc_uncollectable(). In   the  end, this is really
 one of the object-types we want to leave to GC.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef __WINDOWS__
-#define isatty(fd) win_isatty(fd)
-static int
-win_isatty(int fd)
-{ HANDLE h = (HANDLE)_get_osfhandle(fd);
-
-  return GetFileType(h) == FILE_TYPE_CHAR;
-}
-#endif
-
 #ifndef FD_CLOEXEC			/* This is not defined in MacOS */
 #define FD_CLOEXEC 1
 #endif
@@ -3361,7 +3393,11 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   if ( s->fileno >= 0 )
   { int fd = (int)s->fileno;
 
+#ifdef __WINDOWS__
+    if ( win_isconsole(s) )
+#else
     if ( isatty(fd) )
+#endif
       s->flags |= SIO_ISATTY;
 
 #if defined(F_SETFD)
@@ -3696,8 +3732,15 @@ Sfileno(IOSTREAM *s)
 
 HANDLE
 Swinhandle(IOSTREAM *s)
-{ int fd = Sfileno(s);
+{ HANDLE h = NULL;
 
+  if ( s->functions->control &&
+       (*s->functions->control)(s->handle,
+				SIO_GETWINHANDLE,
+				(void *)&h) == 0 )
+    return h;
+
+  int fd = Sfileno(s);
   if ( fd >= 0 )
     return (HANDLE)_get_osfhandle(fd);
 
@@ -4293,7 +4336,17 @@ static const IOSTREAM S__iob0[] =
 };
 
 
-static int S__initialised = false;
+static bool S__initialised = false;
+
+#ifdef __WINDOWS__
+#define isatty(fd) win_isatty(fd)
+static bool
+win_isatty(int fd)
+{ HANDLE h = (HANDLE)_get_osfhandle(fd);
+  DWORD mode;
+  return GetConsoleMode(h, &mode);
+}
+#endif
 
 void
 SinitStreams(void)
