@@ -52,9 +52,13 @@ Event interface
 
 static void	free_event_callback(event_callback *cb);
 
-typedef int (*ev_func0)(void *);
-typedef int (*ev_func1)(void *, term_t);
-typedef int (*ev_func2)(void *, term_t, term_t);
+typedef bool (*ev_func0)(void *);
+typedef bool (*ev_func1)(void *, term_t);
+typedef bool (*ev_func2)(void *, term_t, term_t);
+
+typedef void (*ev_func0v)(void *);
+typedef void (*ev_func1v)(void *, term_t);
+typedef void (*ev_func2v)(void *, term_t, term_t);
 
 #ifdef O_PLMT
 #define INIT_LIST_LOCK(l) recursiveMutexInit(&(l)->lock)
@@ -91,8 +95,8 @@ const event_type PL_events[] =
   {0}
 };
 
-static int
-link_event(event_list *list, event_callback *cb, int last)
+static bool
+link_event(event_list *list, event_callback *cb, bool last)
 { LOCK_LIST(list);
   if ( !list->head )
   { list->head = list->tail = cb;
@@ -121,8 +125,9 @@ get_callback(DECL_LD term_t closure, Module *m, term_t cb)
 }
 
 
-static int
-add_event_hook(event_list *list, atom_t name, int last, term_t closure, int argc)
+static bool
+add_event_hook(event_list *list, atom_t name,
+	       bool last, term_t closure, int argc)
 { GET_LD
   Module m = NULL;
   event_callback *cb;
@@ -189,8 +194,9 @@ get_event_list(event_list **list)
   return *list;
 }
 
-int
-register_event_hook(event_list **list, atom_t name, int last, term_t closure, int argc)
+bool
+register_event_hook(event_list **list, atom_t name,
+		    bool last, term_t closure, int argc)
 { return add_event_hook(get_event_list(list), name, last, closure, argc);
 }
 
@@ -338,12 +344,14 @@ PRED_IMPL("prolog_unlisten", 2, prolog_unlisten, 0)
 }
 
 
-int
-register_event_function(event_list **list, atom_t name, int last, int (*func)(),
-			void *closure, int argc)
+bool
+register_event_function(event_list **list, atom_t name,
+			bool last, void *func,
+			void *closure, int argc, unsigned int flags)
 { event_callback *cb = PL_malloc(sizeof(*cb));
   memset(cb, 0, sizeof(*cb));
   cb->argc = argc;
+  cb->flags = flags;
   cb->function = func;
   cb->closure.pointer = closure;
 
@@ -390,10 +398,12 @@ cleanupEvents(void)
 
 
 
-#define call_event_list(list, argc, argv) LDFUNC(call_event_list, list, argc, argv)
-static int
+#define call_event_list(list, argc, argv) \
+	LDFUNC(call_event_list, list, argc, argv)
+
+static bool
 call_event_list(DECL_LD event_list *list, int argc, term_t argv)
-{ int rc = true;
+{ bool rc = true;
 
   if ( list )
   { event_callback *ev;
@@ -401,19 +411,39 @@ call_event_list(DECL_LD event_list *list, int argc, term_t argv)
     LOCK_LIST(list);
     for(ev = list->head; ev; ev = ev->next)
     { if ( ev->function )
-      { switch(argc)
-	{ case 0:
-	    rc = (*(ev_func0)ev->function)(ev->closure.pointer);
-	    break;
-	  case 1:
-	    rc = (*(ev_func1)ev->function)(ev->closure.pointer, argv+1);
-	    break;
-	  case 2:
-	    rc = (*(ev_func2)ev->function)(ev->closure.pointer, argv+1, argv+2);
-	    break;
-	  default:
+      { if ( ison(ev, EV_CALLBACK_VOID) )
+	{ assert(argc >= ev->argc);
+	  switch(ev->argc)
+	  { case 0:
+	      (*(ev_func0v)ev->function)(ev->closure.pointer);
+	      break;
+	    case 1:
+	      (*(ev_func1v)ev->function)(ev->closure.pointer, argv+1);
+	      break;
+	    case 2:
+	      (*(ev_func2v)ev->function)(ev->closure.pointer, argv+1, argv+2);
+	      break;
+	    default:
+	      rc = false;
+	      assert(0);
+	  }
+	  if ( PL_exception(0) )
 	    rc = false;
-	    assert(0);
+	} else
+	{ switch(argc)
+	  { case 0:
+	      rc = (*(ev_func0)ev->function)(ev->closure.pointer);
+	      break;
+	    case 1:
+	      rc = (*(ev_func1)ev->function)(ev->closure.pointer, argv+1);
+	      break;
+	    case 2:
+	      rc = (*(ev_func2)ev->function)(ev->closure.pointer, argv+1, argv+2);
+	      break;
+	    default:
+	      rc = false;
+	      assert(0);
+	  }
 	}
       } else if ( ev->closure.term )
       { rc = rc &&
@@ -569,7 +599,7 @@ int
 PL_call_event_hook(pl_event_type ev, ...)
 { event_list **listp = event_list_location(ev);
 
-  if ( *listp && GD->cleaning != CLN_DATA )
+  if ( *listp && GD->halt.cleaning != CLN_DATA )
   { va_list args;
     int rc;
 
@@ -588,11 +618,10 @@ PL_call_event_hook(pl_event_type ev, ...)
  * hook
  */
 
-int
-PL_call_event_hook_va(pl_event_type ev, va_list args)
-{ GET_LD
-  wakeup_state wstate;
-  int rc = true;
+bool
+PL_call_event_hook_va(DECL_LD pl_event_type ev, va_list args)
+{ wakeup_state wstate;
+  bool rc = true;
   event_list *list = *event_list_location(ev);
   term_t av;
   const event_type *event_decl = &PL_events[ev];
