@@ -114,7 +114,11 @@ gmp_too_big(DECL_LD gmp_tb why)
       max.value.i = LD->gmp.max_integer_size;
       PL_error(NULL, 0, "requires more than max_integer_size bytes",
 	       ERR_AR_TRIPWIRE, ATOM_max_integer_size, &max);
+#if O_THROW
       PL_rethrow();
+#else
+      PL_fatal_error("Big number retraint not supported without PL_throw()");
+#endif
     }
     case GMP_TB_STACK:
       outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_THROW);
@@ -122,11 +126,19 @@ gmp_too_big(DECL_LD gmp_tb why)
     case GMP_TB_MALLOC:
     default:
       PL_no_memory();
+#if O_THROW
       PL_rethrow();
+#else
+      PL_fatal_error("Too large big number");
+#endif
   }
 
   abortProlog();		/* Just in case the above fails */
+#if O_THROW
   PL_rethrow();
+#else
+  PL_fatal_error("Too large big number");
+#endif
   return false;
 }
 
@@ -172,7 +184,7 @@ mp_alloc(size_t bytes)
   ctx = LD->gmp.context;
 
   size_t fastunits = ROUND_SIZE(bytes)+1;
-  if ( ctx->allocated+fastunits <= GMP_STACK_ALLOC )
+  if ( ctx->allocated+fastunits+1 <= GMP_STACK_ALLOC )
   { size_t *data = &ctx->alloc_buf[ctx->allocated];
     *data++ = fastunits;
     ctx->allocated += fastunits;
@@ -200,7 +212,8 @@ mp_alloc(size_t bytes)
 
 static inline size_t *
 mp_on_stack(ar_context *ctx, void *ptr)
-{ if ( ptr > (void*)ctx->alloc_buf && ptr < (void*)&ctx->alloc_buf[GMP_STACK_ALLOC] )
+{ if ( ptr > (void*)ctx->alloc_buf &&
+       ptr < (void*)&ctx->alloc_buf[GMP_STACK_ALLOC] )
     return &((size_t*)ptr)[-1];
 
   return NULL;
@@ -1286,7 +1299,7 @@ bignum_index(const word *p)
 
 #if O_GMP
 
-int
+bool
 mpz_to_int64(mpz_t mpz, int64_t *i)
 { if ( mpz_cmp(mpz, MPZ_MIN_INT64) >= 0 &&
        mpz_cmp(mpz, MPZ_MAX_INT64) <= 0 )
@@ -1335,7 +1348,7 @@ mpz_to_uint64(mpz_t mpz, uint64_t *i)
 
 #elif O_BF
 
-int
+bool
 mpz_to_int64(mpz_t mpz, int64_t *i)
 { return bf_get_int64(i, mpz, 0) == 0;
 }
@@ -1385,12 +1398,8 @@ put_mpz(DECL_LD Word at, mpz_t mpz, int flags)
 #endif
   { long v = mpz_get_si(mpz);
 
-    if ( !hasGlobalSpace(0) )		/* ensure we have room for bindConst */
-    { int rc = ensureGlobalSpace(0, flags);
-
-      if ( rc != true )
-	return rc;
-    }
+    if ( !ensureGlobalSpace(0, flags) )
+      return false;
 
     *at = consInt(v);
     assert(valInt(*at) == v);
