@@ -75,60 +75,43 @@ clearTermAgenda(term_agenda *a)
 }
 
 
+static inline Word
+nextTermAgendaNoDeRef(term_agenda *a)
+{ if ( unlikely(a->work.size == 0) &&
+       !popSegStack(&a->stack, &a->work, aNode) )
+    return NULL;
+
+  a->work.size--;
+  return a->work.location++;
+}
+
+
 #define nextTermAgenda(a) LDFUNC(nextTermAgenda, a)
 static inline Word
 nextTermAgenda(DECL_LD term_agenda *a)
-{ Word p;
+{ Word p = nextTermAgendaNoDeRef(a);
 
-  if ( a->work.size > 0 )
-  { ok:
-    a->work.size--;
-    p = a->work.location++;
+  if ( p )
     deRef(p);
 
-    return p;
-  }
-
-  if ( popSegStack(&a->stack, &a->work, aNode) )
-    goto ok;
-
-  return NULL;
+  return p;
 }
 
 
-static inline Word
-nextTermAgendaNoDeRef(term_agenda *a)
-{ Word p;
-
-  if ( a->work.size > 0 )
-  { ok:
-    a->work.size--;
-    p = a->work.location++;
-
-    return p;
-  }
-
-  if ( popSegStack(&a->stack, &a->work, aNode) )
-    goto ok;
-
-  return NULL;
-}
-
-
-
-
 		 /*******************************
 		 *	  PUSH VARIATIONS	*
 		 *******************************/
 
 static inline int
 pushWorkAgenda(term_agenda *a, size_t amount, Word start)
-{ if ( a->work.size > 0 )
-  { if ( !pushSegStack(&a->stack, a->work, aNode) )
-      return false;
+{ if ( amount )
+  { if ( a->work.size > 0 )
+    { if ( !pushSegStack(&a->stack, a->work, aNode) )
+	return false;
+    }
+    a->work.location = start;
+    a->work.size = amount;
   }
-  a->work.location = start;
-  a->work.size = amount;
 
   return true;
 }
@@ -281,7 +264,7 @@ nextTermAgenda_P(DECL_LD term_agenda_P *a)
 		 *	  PUSH VARIATIONS	*
 		 *******************************/
 
-static inline int
+static inline bool
 pushWorkAgenda_P(term_agenda_P *a, size_t amount, Word start)
 { if ( a->work.size > 0 )
   { if ( !pushSegStack(&a->stack, a->work, aNode_P) )
@@ -296,6 +279,18 @@ pushWorkAgenda_P(term_agenda_P *a, size_t amount, Word start)
   return true;
 }
 
+static inline bool		/* Avoid last-argument POP merge */
+pushWorkAgenda_P0(term_agenda_P *a, size_t amount, Word start)
+{ if ( !pushSegStack(&a->stack, a->work, aNode_P) )
+    return false;
+  a->work.depth = 1;
+
+  a->work.location = start;
+  a->work.size     = amount;
+
+  return true;
+}
+
 #endif /*!AC_TERM_WALK_POP*/
 
 #if AC_TERM_WALK_LR
