@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2022, SWI-Prolog Solutions b.v.
+    Copyright (c)  2022-2925, SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -67,7 +67,7 @@ write_undo_ref(IOSTREAM *s, atom_t aref, int flags)
   (void)flags;
 
   Sfprintf(s, "<undo>(%p)", r);
-  return TRUE;
+  return true;
 }
 
 static void
@@ -81,7 +81,7 @@ release_undo_blob(atom_t aref)
   record_t r = *ref;
 
   PL_erase(r);
-  return TRUE;
+  return true;
 }
 
 static int
@@ -121,7 +121,7 @@ save_undo(term_t goal)
   return lookupBlob((const char*)&r, sizeof(r), &undo_blob, &new);
 }
 
-static int
+static bool
 put_undo(term_t t, atom_t u)
 { record_t *ref = PL_blob_data(u, NULL, NULL);
   record_t r = *ref;
@@ -149,9 +149,9 @@ PRED_IMPL("$undo", 1, undo, 0)
       DEBUG(MSG_UNDO, Sdprintf("Trailed %p\n", ul));
       *ul = consPtr(p, TAG_COMPOUND|STG_GLOBAL);
 
-      return TRUE;
+      return true;
     } else
-      return FALSE;
+      return false;
   } else
     return PL_permission_error("undo", "goal", A1);
 }
@@ -207,7 +207,7 @@ put_scheduled_undo(DECL_LD term_t list)
 
       if ( !put_undo(h, u) ||
 	   !PL_cons_list(list, h, list) )
-	return FALSE;
+	return false;
     }
 
     for(; top0 > top; )
@@ -217,17 +217,17 @@ put_scheduled_undo(DECL_LD term_t list)
     emptyBuffer(LD->undo.scheduled, 1024);
   }
 
-  return TRUE;
+  return true;
 }
 
-int
+bool
 run_undo_hooks(DECL_LD)
 { fid_t fid;
 
   if ( (fid=PL_open_foreign_frame()) )
   { predicate_t pred;
     term_t list;
-    int rc;
+    bool rc;
 
     pred = _PL_predicate("$run_undo", 1, "$syspreds", &GD->procedures.drun_undo1);
 
@@ -241,7 +241,7 @@ run_undo_hooks(DECL_LD)
     return rc;
   }
 
-  return FALSE;
+  return false;
 }
 
 
